@@ -45,9 +45,9 @@
 
 static const char *TE133_FW_FILENAME = "dahdi-fw-te133.bin";
 static const char *TE134_FW_FILENAME = "dahdi-fw-te134.bin";
-static const u32 TE13X_FW_VERSION = 0x6f0017;
+static const u32 TE133_FW_VERSION = 0x780019;
+static const u32 TE134_FW_VERSION = 0x780017;
 
-#define VPM_SUPPORT
 #define WC_MAX_IFACES 8
 
 enum linemode {
@@ -73,7 +73,6 @@ enum linemode {
 #define FALC_INT (1<<3)
 
 struct t13x {
-	struct pci_dev *dev;
 	spinlock_t reglock;
 
 	u8 latency;
@@ -93,9 +92,10 @@ struct t13x {
 	unsigned long loopuptimer;
 	unsigned long loopdntimer;
 	const char *name;
-#define INITIALIZED    1
-#define SHUTDOWN       2
-#define READY	       3
+#define INITIALIZED		1
+#define SHUTDOWN		2
+#define READY			3
+#define HAVE_OPEN_CHANNELS	4
 	unsigned long bit_flags;
 	u32 ledstate;
 	struct dahdi_device *ddev;
@@ -107,21 +107,27 @@ struct t13x {
 	struct timer_list timer;
 	struct work_struct timer_work;
 	struct workqueue_struct *wq;
-#ifdef VPM_SUPPORT
 	struct vpm450m *vpm;
-#endif
 	struct mutex lock;
 	struct wcxb xb;
+	u32    lineconfig_fingerprint;
 };
 
 static void te13x_handle_transmit(struct wcxb *xb, void *vfp);
 static void te13x_handle_receive(struct wcxb *xb, void *vfp);
 static void te13x_handle_interrupt(struct wcxb *xb, u32 pending);
 
+static void te13x_handle_error(struct wcxb *xb)
+{
+	struct t13x *wc = container_of(xb, struct t13x, xb);
+	wc->ddev->irqmisses++;
+}
+
 static struct wcxb_operations xb_ops = {
 	.handle_receive = te13x_handle_receive,
 	.handle_transmit = te13x_handle_transmit,
 	.handle_interrupt = te13x_handle_interrupt,
+	.handle_error = te13x_handle_error,
 };
 
 /* Maintenance Mode Registers */
@@ -178,6 +184,7 @@ static int yelalarmdebounce = 500; /* RAI(yellow) def to 0.5s AT&T devguide */
 static char *default_linemode = "t1"; /* 'e1', 't1', or 'j1' */
 static int force_firmware;
 static int latency = WCXB_DEFAULT_LATENCY;
+static unsigned int max_latency = WCXB_DEFAULT_MAXLATENCY;
 
 struct t13x_firm_header {
 	u8	header[6];
@@ -190,7 +197,6 @@ struct t13x_firm_header {
 static void t13x_check_alarms(struct t13x *wc);
 static void t13x_check_sigbits(struct t13x *wc);
 
-#ifdef VPM_SUPPORT
 #include <linux/slab.h>
 #include <linux/vmalloc.h>
 #include <linux/string.h>
@@ -533,23 +539,23 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 
 	vpm450m = kzalloc(sizeof(struct vpm450m), GFP_KERNEL);
 	if (!vpm450m) {
-		dev_info(&wc->dev->dev, "Unable to allocate vpm450m struct\n");
+		dev_info(&wc->xb.pdev->dev, "Unable to allocate vpm450m struct\n");
 		return NULL;
 	}
 
-	vpm450m->context.dev = &wc->dev->dev;
+	vpm450m->context.dev = &wc->xb.pdev->dev;
 	vpm450m->context.ops = &t13x_oct612x_ops;
 
 	ChipOpen = kzalloc(sizeof(tOCT6100_CHIP_OPEN), GFP_KERNEL);
 	if (!ChipOpen) {
-		dev_info(&wc->dev->dev, "Unable to allocate ChipOpen\n");
+		dev_info(&wc->xb.pdev->dev, "Unable to allocate ChipOpen\n");
 		kfree(vpm450m);
 		return NULL;
 	}
 
 	ChannelOpen = kzalloc(sizeof(tOCT6100_CHANNEL_OPEN), GFP_KERNEL);
 	if (!ChannelOpen) {
-		dev_info(&wc->dev->dev, "Unable to allocate ChannelOpen\n");
+		dev_info(&wc->xb.pdev->dev, "Unable to allocate ChannelOpen\n");
 		kfree(vpm450m);
 		kfree(ChipOpen);
 		return NULL;
@@ -558,7 +564,7 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 	for (x = 0; x < ARRAY_SIZE(vpm450m->ecmode); x++)
 		vpm450m->ecmode[x] = -1;
 
-	dev_info(&wc->dev->dev, "Echo cancellation for %d channels\n",
+	dev_info(&wc->xb.pdev->dev, "Echo cancellation for %d channels\n",
 		 ECHOCAN_NUM_CHANS);
 
 	Oct6100ChipOpenDef(ChipOpen);
@@ -620,14 +626,14 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 	/* Get the size of the OCT6100 instance structure. */
 	ulResult = Oct6100GetInstanceSize(ChipOpen, &InstanceSize);
 	if (ulResult != cOCT6100_ERR_OK) {
-		dev_info(&wc->dev->dev, "Unable to get instance size: %x\n",
+		dev_info(&wc->xb.pdev->dev, "Unable to get instance size: %x\n",
 				ulResult);
 		return NULL;
 	}
 
 	vpm450m->pApiInstance = vmalloc(InstanceSize.ulApiInstanceSize);
 	if (!vpm450m->pApiInstance) {
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			"Out of memory (can't allocate %d bytes)!\n",
 			InstanceSize.ulApiInstanceSize);
 		return NULL;
@@ -637,7 +643,7 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 	oct_enable_dram(wc);
 	ulResult = Oct6100ChipOpen(vpm450m->pApiInstance, ChipOpen);
 	if (ulResult != cOCT6100_ERR_OK) {
-		dev_info(&wc->dev->dev, "Unable to Oct6100ChipOpen: %x\n",
+		dev_info(&wc->xb.pdev->dev, "Unable to Oct6100ChipOpen: %x\n",
 				ulResult);
 		return NULL;
 	}
@@ -647,7 +653,7 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 	for (i = 0; i < ECHOCAN_NUM_CHANS; i++) {
 		ulResult = echocan_initialize_channel(vpm450m, i, isalaw);
 		if (0 != ulResult) {
-			dev_info(&wc->dev->dev,
+			dev_info(&wc->xb.pdev->dev,
 				"Unable to echocan_initialize_channel: %x\n",
 				ulResult);
 			return NULL;
@@ -656,7 +662,7 @@ static struct vpm450m *init_vpm450m(struct t13x *wc, int isalaw,
 		}
 	}
 
-	if (vpmsupport != 0)
+	if (vpmsupport)
 		wcxb_enable_echocan(&wc->xb);
 
 	kfree(ChipOpen);
@@ -709,7 +715,7 @@ static int t13x_echocan_create(struct dahdi_chan *chan,
 	features = &vpm_ec_features;
 
 	if (ecp->param_count > 0) {
-		dev_warn(&wc->dev->dev, "%s echo canceller does not support parameters; failing request\n",
+		dev_warn(&wc->xb.pdev->dev, "%s echo canceller does not support parameters; failing request\n",
 			 chan->ec_factory->get_name(chan));
 		return -EINVAL;
 	}
@@ -747,14 +753,14 @@ static void t13x_vpm_init(struct t13x *wc)
 #endif
 
 	if (!vpmsupport) {
-		dev_info(&wc->dev->dev, "VPM450: Support Disabled\n");
+		dev_info(&wc->xb.pdev->dev, "VPM450: Support Disabled\n");
 		return;
 	}
 
 #if defined(HOTPLUG_FIRMWARE)
-	if ((request_firmware(&firmware, oct032_firmware, &wc->dev->dev) != 0)
-			|| !firmware) {
-		dev_notice(&wc->dev->dev, "VPM450: firmware %s not available from userspace\n",
+	if ((request_firmware(&firmware, oct032_firmware,
+			      &wc->xb.pdev->dev) != 0) || !firmware) {
+		dev_notice(&wc->xb.pdev->dev, "VPM450: firmware %s not available from userspace\n",
 				oct032_firmware);
 		return;
 	}
@@ -773,7 +779,7 @@ static void t13x_vpm_init(struct t13x *wc)
 
 	wc->vpm = init_vpm450m(wc, companding, firmware);
 	if (!wc->vpm) {
-		dev_notice(&wc->dev->dev, "VPM450: Failed to initialize\n");
+		dev_notice(&wc->xb.pdev->dev, "VPM450: Failed to initialize\n");
 		if (firmware != &embedded_firmware)
 			release_firmware(firmware);
 		return;
@@ -782,11 +788,10 @@ static void t13x_vpm_init(struct t13x *wc)
 	if (firmware != &embedded_firmware)
 		release_firmware(firmware);
 
-	dev_info(&wc->dev->dev,
+	dev_info(&wc->xb.pdev->dev,
 		"VPM450: Present and operational servicing %d span\n", 1);
 
 }
-#endif /* VPM_SUPPORT */
 
 static int t13x_clear_maint(struct dahdi_span *span);
 
@@ -796,8 +801,8 @@ struct t13x_desc {
 	const char *name;
 };
 
-static const struct t13x_desc te133 = {"Wildcard TE133"}; /* pci express */
-static const struct t13x_desc te134 = {"Wildcard TE134"}; /* legacy pci */
+static const struct t13x_desc te133 = {"Wildcard TE131/TE133"}; /* pci express*/
+static const struct t13x_desc te134 = {"Wildcard TE132/TE134"}; /* legacy pci */
 
 static inline bool is_pcie(const struct t13x *t1)
 {
@@ -1134,7 +1139,7 @@ static void t13x_configure_t1(struct t13x *wc, int lineconfig, int txlevel)
 	__t13x_framer_set(wc, 0x18, 0x3f);	/* IMR4: Slips on transmit */
 
 	spin_unlock_irqrestore(&wc->reglock, flags);
-	dev_info(&wc->dev->dev, "Span configured for %s/%s\n", framing, line);
+	dev_info(&wc->xb.pdev->dev, "Span configured for %s/%s\n", framing, line);
 }
 
 static void t13x_configure_e1(struct t13x *wc, int lineconfig)
@@ -1233,7 +1238,7 @@ static void t13x_configure_e1(struct t13x *wc, int lineconfig)
 							   transmit */
 
 	spin_unlock_irqrestore(&wc->reglock, flags);
-	dev_info(&wc->dev->dev,
+	dev_info(&wc->xb.pdev->dev,
 			"Span configured for %s/%s%s\n", framing, line, crc4);
 }
 
@@ -1251,25 +1256,6 @@ static void t13x_framer_start(struct t13x *wc)
 	set_bit(DAHDI_FLAGBIT_RUNNING, &wc->span.flags);
 }
 
-static void set_span_devicetype(struct t13x *wc)
-{
-	const char *olddevicetype;
-	olddevicetype = wc->ddev->devicetype;
-
-#ifndef VPM_SUPPORT
-	wc->ddev->devicetype = kasprintf(GFP_KERNEL,
-				 "%s (VPMOCT032)", wc->devtype->name);
-#else
-	wc->ddev->devicetype = kasprintf(GFP_KERNEL, "%s", wc->devtype->name);
-#endif /* VPM_SUPPORT */
-
-	/* On the off chance that we were able to allocate it previously. */
-	if (!wc->ddev->devicetype)
-		wc->ddev->devicetype = olddevicetype;
-	else
-		kfree(olddevicetype);
-}
-
 /**
  * te13xp_check_for_interrupts - Return 0 if the card is generating interrupts.
  * @wc:	The card to check.
@@ -1290,7 +1276,7 @@ static int te13xp_check_for_interrupts(struct t13x *wc)
 		spin_unlock_irqrestore(&wc->reglock, flags);
 		if (time_after(jiffies, stop_time)) {
 			wc->span.alarms = DAHDI_ALARM_RED;
-			dev_err(&wc->dev->dev, "Interrupts not detected.\n");
+			dev_err(&wc->xb.pdev->dev, "Interrupts not detected.\n");
 			return -EIO;
 		}
 		msleep(100);
@@ -1301,25 +1287,39 @@ static int te13xp_check_for_interrupts(struct t13x *wc)
 	return 0;
 }
 
+static bool have_open_channels(const struct t13x *wc)
+{
+	int x, j;
+	for (x = 0, j = 0; x < wc->span.channels; x++) {
+		const struct dahdi_chan *chan = wc->span.chans[x];
+		if (test_bit(DAHDI_FLAGBIT_OPEN, &chan->flags) ||
+		    dahdi_have_netdev(chan))
+			return true;
+	}
+	return false;
+}
+
 static int t13x_startup(struct file *file, struct dahdi_span *span)
 {
 	struct t13x *wc = container_of(span, struct t13x, span);
 	unsigned long flags;
 	int ret;
 
-	set_span_devicetype(wc);
-
 	/* Stop the DMA since the clock source may have changed. */
 	wcxb_stop_dma(&wc->xb);
 	ret = wcxb_wait_for_stop(&wc->xb, 50);
 	if (ret) {
-		dev_err(&wc->dev->dev, "Timeout waiting for DMA to stop.\n");
+		dev_err(&wc->xb.pdev->dev, "Timeout waiting for DMA to stop.\n");
 		return ret;
 	}
 
 	/* Reset framer with proper parameters and start */
 	t13x_framer_start(wc);
 
+	/* Reset span alarm state to RED to prevent false
+	 * temporary GREEN state on span bringup */
+	span->alarms |= DAHDI_ALARM_RED;
+
 	/* Do we want to SYNC on receive or not. This must always happen after
 	 * the framer is fully reset. */
 	wcxb_set_clksrc(&wc->xb,
@@ -1330,9 +1330,15 @@ static int t13x_startup(struct file *file, struct dahdi_span *span)
 	if (ret)
 		return ret;
 
-	dev_info(&wc->dev->dev,
+	dev_info(&wc->xb.pdev->dev,
 			"Calling startup (flags is %lu)\n", span->flags);
 
+	/* Check for "open channels" here in case some channels have netdev. */
+	if (have_open_channels(wc))
+		clear_bit(HAVE_OPEN_CHANNELS, &wc->bit_flags);
+	else
+		set_bit(HAVE_OPEN_CHANNELS, &wc->bit_flags);
+
 	/* Get this party started */
 	local_irq_save(flags);
 	t13x_check_alarms(wc);
@@ -1491,11 +1497,11 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 
 	switch (cmd) {
 	case DAHDI_MAINT_NONE:
-		dev_info(&wc->dev->dev, "Clearing all maint modes\n");
+		dev_info(&wc->xb.pdev->dev, "Clearing all maint modes\n");
 		t13x_clear_maint(span);
 		break;
 	case DAHDI_MAINT_LOCALLOOP:
-		dev_info(&wc->dev->dev, "Turning on local loopback\n");
+		dev_info(&wc->xb.pdev->dev, "Turning on local loopback\n");
 		t13x_clear_maint(span);
 		spin_lock_irqsave(&wc->reglock, flags);
 		reg = __t13x_framer_get(wc, LIM0);
@@ -1503,7 +1509,7 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 		spin_unlock_irqrestore(&wc->reglock, flags);
 		break;
 	case DAHDI_MAINT_NETWORKLINELOOP:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 				"Turning on network line loopback\n");
 		t13x_clear_maint(span);
 		spin_lock_irqsave(&wc->reglock, flags);
@@ -1512,7 +1518,7 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 		spin_unlock_irqrestore(&wc->reglock, flags);
 		break;
 	case DAHDI_MAINT_NETWORKPAYLOADLOOP:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			"Turning on network payload loopback\n");
 		t13x_clear_maint(span);
 		spin_lock_irqsave(&wc->reglock, flags);
@@ -1521,7 +1527,7 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 		spin_unlock_irqrestore(&wc->reglock, flags);
 		break;
 	case DAHDI_MAINT_LOOPUP:
-		dev_info(&wc->dev->dev, "Transmitting loopup code\n");
+		dev_info(&wc->xb.pdev->dev, "Transmitting loopup code\n");
 		t13x_clear_maint(span);
 		if (dahdi_is_e1_span(&wc->span)) {
 			spin_lock_irqsave(&wc->reglock, flags);
@@ -1536,7 +1542,7 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 		}
 		break;
 	case DAHDI_MAINT_LOOPDOWN:
-		dev_info(&wc->dev->dev, "Transmitting loopdown code\n");
+		dev_info(&wc->xb.pdev->dev, "Transmitting loopdown code\n");
 		t13x_clear_maint(span);
 		if (dahdi_is_e1_span(&wc->span)) {
 			spin_lock_irqsave(&wc->reglock, flags);
@@ -1572,7 +1578,7 @@ static int t13x_maint(struct dahdi_span *span, int cmd)
 		t13x_reset_counters(span);
 		break;
 	default:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 				"Unknown maint command: %d\n", cmd);
 		return -ENOSYS;
 	}
@@ -1674,25 +1680,56 @@ static void t13x_chan_set_sigcap(struct dahdi_span *span, int x)
 	}
 }
 
+static bool t13x_lineconfig_changed(struct t13x *wc,
+				    const struct dahdi_lineconfig *lc)
+{
+	unsigned long flags;
+	bool result;
+	u32 crc = crc32(~0, lc, sizeof(*lc));
+
+	spin_lock_irqsave(&wc->reglock, flags);
+	result = (crc != wc->lineconfig_fingerprint);
+	spin_unlock_irqrestore(&wc->reglock, flags);
+	return result;
+}
+
+static void t13x_save_lineconfig(struct t13x *wc,
+				 const struct dahdi_lineconfig *lc)
+{
+	unsigned long flags;
+	u32 crc = crc32(~0, lc, sizeof(*lc));
+
+	spin_lock_irqsave(&wc->reglock, flags);
+	wc->lineconfig_fingerprint = crc;
+	spin_unlock_irqrestore(&wc->reglock, flags);
+}
+
 static int
 t13x_spanconfig(struct file *file, struct dahdi_span *span,
 		 struct dahdi_lineconfig *lc)
 {
 	struct t13x *wc = container_of(span, struct t13x, span);
 	int i;
+	int res = 0;
 
 	if (file->f_flags & O_NONBLOCK)
 		return -EAGAIN;
 
+	/* Set recover timing mode */
+	span->syncsrc = lc->sync;
+
 	/* make sure that sigcaps gets updated if necessary */
 	for (i = 0; i < wc->span.channels; i++)
 		t13x_chan_set_sigcap(span, i);
 
 	/* If already running, apply changes immediately */
-	if (test_bit(DAHDI_FLAGBIT_RUNNING, &span->flags))
-		return t13x_startup(file, span);
-
-	return 0;
+	if (test_bit(DAHDI_FLAGBIT_RUNNING, &span->flags) &&
+	    t13x_lineconfig_changed(wc, lc)) {
+		res = t13x_startup(file, span);
+		if (!res)
+			t13x_save_lineconfig(wc, lc);
+	}
+	return res;
 }
 
 /**
@@ -1723,8 +1760,6 @@ static int t13x_software_init(struct t13x *wc, enum linemode type)
 			goto error_exit;
 	}
 
-	set_span_devicetype(wc);
-
 	/* Because the interrupt handler is running, we need to atomically
 	 * swap the channel arrays. */
 	spin_lock_irqsave(&wc->reglock, flags);
@@ -1764,7 +1799,7 @@ static int t13x_software_init(struct t13x *wc, enum linemode type)
 
 	spin_unlock_irqrestore(&wc->reglock, flags);
 
-	dev_info(&wc->dev->dev, "Setting up global serial parameters for %s\n",
+	dev_info(&wc->xb.pdev->dev, "Setting up global serial parameters for %s\n",
 		dahdi_spantype2str(wc->span.spantype));
 
 	t13x_serial_setup(wc);
@@ -1776,8 +1811,6 @@ static int t13x_software_init(struct t13x *wc, enum linemode type)
 		wc->chans[x]->chanpos = x + 1;
 	}
 
-	set_span_devicetype(wc);
-
 	return 0;
 
 error_exit:
@@ -1816,7 +1849,6 @@ static int t13x_set_linemode(struct dahdi_span *span, enum spantypes linemode)
 	clear_bit(INITIALIZED, &wc->bit_flags);
 	disable_irq(wc->xb.pdev->irq);
 
-	synchronize_irq(wc->dev->irq);
 	smp_mb__after_clear_bit();
 	del_timer_sync(&wc->timer);
 	flush_workqueue(wc->wq);
@@ -1826,24 +1858,24 @@ static int t13x_set_linemode(struct dahdi_span *span, enum spantypes linemode)
 
 	switch (linemode) {
 	case SPANTYPE_DIGITAL_T1:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			 "Changing from %s to T1 line mode.\n",
 			 dahdi_spantype2str(wc->span.spantype));
 		res = t13x_software_init(wc, T1);
 		break;
 	case SPANTYPE_DIGITAL_E1:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			 "Changing from %s to E1 line mode.\n",
 			 dahdi_spantype2str(wc->span.spantype));
 		res = t13x_software_init(wc, E1);
 		break;
 	case SPANTYPE_DIGITAL_J1:
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			 "Changing from %s to E1 line mode.\n",
 			 dahdi_spantype2str(wc->span.spantype));
 		res = t13x_software_init(wc, J1);
 	default:
-		dev_err(&wc->dev->dev,
+		dev_err(&wc->xb.pdev->dev,
 			"Got invalid linemode '%s' from dahdi\n",
 			dahdi_spantype2str(linemode));
 		res = -EINVAL;
@@ -1876,14 +1908,14 @@ static int t13x_hardware_post_init(struct t13x *wc, enum linemode *type)
 	} else if (!strcasecmp(default_linemode, "j1")) {
 		*type = J1;
 	} else {
-		dev_warn(&wc->dev->dev,
+		dev_warn(&wc->xb.pdev->dev,
 			 "'%s' is an unknown linemode. Defaulting to 't1'\n",
 			 default_linemode);
 		*type = T1;
 	}
 
 	if (debug) {
-		dev_info(&wc->dev->dev, "linemode: %s\n",
+		dev_info(&wc->xb.pdev->dev, "linemode: %s\n",
 			 (*type == T1) ? "T1" : ((J1 == *type) ? "J1" : "E1"));
 	}
 
@@ -1891,23 +1923,23 @@ static int t13x_hardware_post_init(struct t13x *wc, enum linemode *type)
 	wcxb_gpio_set(&wc->xb, FALC_CPU_RESET);
 	reg = t13x_framer_get(wc, 0x4a);
 	if (reg < 0) {
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 				"Failed to read FALC version (%x)\n", reg);
 		return -EIO;
 	}
-	dev_info(&wc->dev->dev, "FALC version: %1x\n", reg);
+	dev_info(&wc->xb.pdev->dev, "FALC version: %1x\n", reg);
 
 	/* make sure reads and writes work */
 	for (x = 0; x < 256; x++) {
 		t13x_framer_set(wc, 0x14, x);
 		reg = t13x_framer_get(wc, 0x14);
 		if (reg < 0) {
-			dev_info(&wc->dev->dev,
+			dev_info(&wc->xb.pdev->dev,
 					"Failed register read (%d)\n", reg);
 			return -EIO;
 		}
 		if (reg != x) {
-			dev_info(&wc->dev->dev,
+			dev_info(&wc->xb.pdev->dev,
 				"Register test failed. Wrote '%x' but read '%x'\n",
 				x, reg);
 			return -EIO;
@@ -1924,7 +1956,6 @@ static void t13x_check_alarms(struct t13x *wc)
 {
 	unsigned char c, d;
 	int alarms;
-	int x, j;
 
 	if (!(test_bit(DAHDI_FLAGBIT_RUNNING, &wc->span.flags)))
 		return;
@@ -1946,7 +1977,7 @@ static void t13x_check_alarms(struct t13x *wc)
 				/* LIM0: Force RAI High */
 				__t13x_framer_set(wc, 0x20, 0x9f | 0x20);
 				wc->flags.nmf = 1;
-				dev_info(&wc->dev->dev, "NMF workaround on!\n");
+				dev_info(&wc->xb.pdev->dev, "NMF workaround on!\n");
 			}
 			__t13x_framer_set(wc, 0x1e, 0xc3);	/* Reset to CRC4 mode */
 			__t13x_framer_set(wc, 0x1c, 0xf2);	/* Force Resync */
@@ -1956,18 +1987,14 @@ static void t13x_check_alarms(struct t13x *wc)
 				/* LIM0: Clear forced RAI */
 				__t13x_framer_set(wc, 0x20, 0x9f);
 				wc->flags.nmf = 0;
-				dev_info(&wc->dev->dev,
+				dev_info(&wc->xb.pdev->dev,
 						"NMF workaround off!\n");
 			}
 		}
 	}
 
 	if (wc->span.lineconfig & DAHDI_CONFIG_NOTOPEN) {
-		for (x = 0, j = 0; x < wc->span.channels; x++)
-			if ((wc->span.chans[x]->flags & DAHDI_FLAG_OPEN) ||
-			    dahdi_have_netdev(wc->span.chans[x]))
-				j++;
-		if (!j)
+		if (!test_bit(HAVE_OPEN_CHANNELS, &wc->bit_flags))
 			alarms |= DAHDI_ALARM_NOTOPEN;
 		else
 			alarms &= ~DAHDI_ALARM_NOTOPEN;
@@ -2065,31 +2092,31 @@ static void t13x_debounce_alarms(struct t13x *wc)
 	if (wc->lofalarmtimer && time_after(jiffies, wc->lofalarmtimer)) {
 		alarms |= DAHDI_ALARM_RED;
 		wc->lofalarmtimer = 0;
-		dev_info(&wc->dev->dev, "LOF alarm detected\n");
+		dev_info(&wc->xb.pdev->dev, "LOF alarm detected\n");
 	}
 
 	if (wc->losalarmtimer && time_after(jiffies, wc->losalarmtimer)) {
 		alarms |= DAHDI_ALARM_RED;
 		wc->losalarmtimer = 0;
-		dev_info(&wc->dev->dev, "LOS alarm detected\n");
+		dev_info(&wc->xb.pdev->dev, "LOS alarm detected\n");
 	}
 
 	if (wc->aisalarmtimer && time_after(jiffies, wc->aisalarmtimer)) {
 		alarms |= DAHDI_ALARM_BLUE;
 		wc->aisalarmtimer = 0;
-		dev_info(&wc->dev->dev, "AIS alarm detected\n");
+		dev_info(&wc->xb.pdev->dev, "AIS alarm detected\n");
 	}
 
 	if (wc->yelalarmtimer && time_after(jiffies, wc->yelalarmtimer)) {
 		alarms |= DAHDI_ALARM_YELLOW;
 		wc->yelalarmtimer = 0;
-		dev_info(&wc->dev->dev, "YEL alarm detected\n");
+		dev_info(&wc->xb.pdev->dev, "YEL alarm detected\n");
 	}
 
 	if (wc->recoverytimer && time_after(jiffies, wc->recoverytimer)) {
 		alarms &= ~(DAHDI_ALARM_RECOVER);
 		wc->recoverytimer = 0;
-		dev_info(&wc->dev->dev, "Alarms cleared\n");
+		dev_info(&wc->xb.pdev->dev, "Alarms cleared\n");
 	}
 
 	if (alarms != wc->span.alarms) {
@@ -2103,7 +2130,7 @@ static void t13x_debounce_alarms(struct t13x *wc)
 	if (alarms & (DAHDI_ALARM_RED|DAHDI_ALARM_BLUE|
 				DAHDI_ALARM_NOTOPEN|DAHDI_ALARM_RECOVER)) {
 		if (!wc->flags.sendingyellow) {
-			dev_info(&wc->dev->dev, "Setting yellow alarm\n");
+			dev_info(&wc->xb.pdev->dev, "Setting yellow alarm\n");
 			/* We manually do yellow alarm to handle RECOVER
 			 * and NOTOPEN, otherwise it's auto anyway */
 			fmr4 = __t13x_framer_get(wc, 0x20);
@@ -2112,7 +2139,7 @@ static void t13x_debounce_alarms(struct t13x *wc)
 		}
 	} else {
 		if (wc->flags.sendingyellow) {
-			dev_info(&wc->dev->dev, "Clearing yellow alarm\n");
+			dev_info(&wc->xb.pdev->dev, "Clearing yellow alarm\n");
 			/* We manually do yellow alarm to handle RECOVER  */
 			fmr4 = __t13x_framer_get(wc, 0x20);
 			__t13x_framer_set(wc, 0x20, fmr4 & ~0x20);
@@ -2130,7 +2157,7 @@ static void t13x_debounce_loopcodes(struct t13x *wc)
 	spin_lock_irqsave(&wc->reglock, flags);
 	if (wc->loopuptimer && time_after(jiffies, wc->loopuptimer)) {
 		/* Loop-up code debounced */
-		dev_info(&wc->dev->dev, "Loopup detected, enabling remote loop\n");
+		dev_info(&wc->xb.pdev->dev, "Loopup detected, enabling remote loop\n");
 		__t13x_framer_set(wc, 0x36, 0x08);	/* LIM0: Disable
 							   any local loop */
 		__t13x_framer_set(wc, 0x37, 0xf6);	/* LIM1: Enable
@@ -2144,7 +2171,7 @@ static void t13x_debounce_loopcodes(struct t13x *wc)
 
 	if (wc->loopdntimer && time_after(jiffies, wc->loopdntimer)) {
 		/* Loop-down code debounced */
-		dev_info(&wc->dev->dev, "Loopdown detected, disabling remote loop\n");
+		dev_info(&wc->xb.pdev->dev, "Loopdown detected, disabling remote loop\n");
 		__t13x_framer_set(wc, 0x36, 0x08);	/* LIM0: Disable
 							   any local loop */
 		__t13x_framer_set(wc, 0x37, 0xf0);	/* LIM1: Disable
@@ -2208,7 +2235,7 @@ static void te13x_handle_receive(struct wcxb *xb, void *vfp)
 		}
 	}
 
-	if (0 == vpmsupport) {
+	if (!vpmsupport || !wc->vpm) {
 		for (i = 0; i < wc->span.channels; i++) {
 			struct dahdi_chan *const c = wc->span.chans[i];
 			__dahdi_ec_chunk(c, c->readchunk, c->readchunk,
@@ -2256,7 +2283,8 @@ static void timer_work_func(struct work_struct *work)
 	static int work_count;
 
 	if (debug)
-		dev_notice(&wc->dev->dev, "Timer work: %d!\n", ++work_count);
+		dev_notice(&wc->xb.pdev->dev, "Timer work: %d!\n",
+			   ++work_count);
 
 	t13x_debounce_alarms(wc);
 	if (!dahdi_is_e1_span(&wc->span))
@@ -2288,7 +2316,7 @@ static void handle_falc_int(struct t13x *wc)
 	isr4 = (gis & FRMR_GIS_ISR4) ? __t13x_framer_get(wc, FRMR_ISR4) : 0;
 
 	if ((debug) && !(isr3 & ISR3_SEC)) {
-		dev_info(&wc->dev->dev, "gis: %02x, isr0: %02x, isr1: %02x, "\
+		dev_info(&wc->xb.pdev->dev, "gis: %02x, isr0: %02x, isr1: %02x, "\
 			"isr2: %02x, isr3: %02x, isr4: %02x, intcount=%u\n",
 			gis, isr0, isr1, isr2, isr3, isr4, intcount);
 	}
@@ -2367,6 +2395,41 @@ static void te13xp_timer(unsigned long data)
 static inline void create_sysfs_files(struct t13x *wc) { return; }
 static inline void remove_sysfs_files(struct t13x *wc) { return; }
 
+static int t13x_open(struct dahdi_chan *chan)
+{
+	struct t13x *wc = chan->pvt;
+	unsigned long flags;
+
+	if (!(wc->span.lineconfig & DAHDI_CONFIG_NOTOPEN))
+		return 0;
+
+	if (!test_and_set_bit(HAVE_OPEN_CHANNELS, &wc->bit_flags)) {
+		local_irq_save(flags);
+		t13x_check_alarms(wc);
+		local_irq_restore(flags);
+	}
+
+	return 0;
+}
+
+static int t13x_close(struct dahdi_chan *chan)
+{
+	struct t13x *wc = chan->pvt;
+	unsigned long flags;
+
+	if (!(wc->span.lineconfig & DAHDI_CONFIG_NOTOPEN))
+		return 0;
+
+	if (!have_open_channels(wc)) {
+		if (test_and_clear_bit(HAVE_OPEN_CHANNELS, &wc->bit_flags)) {
+			local_irq_save(flags);
+			t13x_check_alarms(wc);
+			local_irq_restore(flags);
+		}
+	}
+	return 0;
+}
+
 static const struct dahdi_span_ops t13x_span_ops = {
 	.owner = THIS_MODULE,
 	.spanconfig = t13x_spanconfig,
@@ -2376,10 +2439,10 @@ static const struct dahdi_span_ops t13x_span_ops = {
 	.maint = t13x_maint,
 	.ioctl = t13x_ioctl,
 	.set_spantype = t13x_set_linemode,
-#ifdef VPM_SUPPORT
 	.echocan_create = t13x_echocan_create,
 	.echocan_name = t13x_echocan_name,
-#endif /* VPM_SUPPORT */
+	.open = t13x_open,
+	.close = t13x_close,
 };
 
 #define SPI_BASE 0x200
@@ -2452,6 +2515,30 @@ error_exit:
 	return serial;
 }
 
+static int te13xp_check_firmware(struct t13x *wc)
+{
+	const char *filename;
+	u32 expected_version;
+	enum wcxb_reset_option reset;
+
+	if (is_pcie(wc)) {
+		filename = TE133_FW_FILENAME;
+		expected_version = TE133_FW_VERSION;
+	} else {
+		filename = TE134_FW_FILENAME;
+		expected_version = TE134_FW_VERSION;
+	}
+
+	/* Specific firmware requires power cycle to properly reset */
+	if (0x6f0017 == wcxb_get_firmware_version(&wc->xb))
+		reset = WCXB_RESET_LATER;
+	else
+		reset = WCXB_RESET_NOW;
+
+	return wcxb_check_firmware(&wc->xb, expected_version, filename,
+			force_firmware, reset);
+}
+
 static int __devinit te13xp_init_one(struct pci_dev *pdev,
 		const struct pci_device_id *ent)
 {
@@ -2461,6 +2548,7 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	int res;
 	unsigned int index = -1;
 	enum linemode type;
+	bool vpmcapable = false;
 
 	for (x = 0; x < ARRAY_SIZE(ifaces); x++) {
 		if (!ifaces[x]) {
@@ -2480,7 +2568,6 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 		return -ENOMEM;
 	}
 
-	wc->dev = pdev;
 	wc->devtype = d;
 	dma_set_mask(&pdev->dev, DMA_BIT_MASK(32));
 
@@ -2529,24 +2616,26 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	wc->xb.pdev = pdev;
 	wc->xb.ops = &xb_ops;
 	wc->xb.debug = &debug;
-	res = wcxb_init(&wc->xb, wc->name, 1);
+	res = wcxb_init(&wc->xb, wc->name, 0);
 	if (res)
 		goto fail_exit;
 
-	/* Check for field updatable firmware */
-	if (is_pcie(wc)) {
-		res = wcxb_check_firmware(&wc->xb, TE13X_FW_VERSION,
-				TE133_FW_FILENAME, force_firmware);
-		if (res)
-			goto fail_exit;
-	} else {
-		res = wcxb_check_firmware(&wc->xb, TE13X_FW_VERSION,
-				TE134_FW_FILENAME, force_firmware);
-		if (res)
-			goto fail_exit;
-	}
+	res = te13xp_check_firmware(wc);
+	if (res)
+		goto fail_exit;
 
 	wc->ddev->hardware_id = t13x_read_serial(wc);
+	if (!wc->ddev->hardware_id) {
+		dev_info(&wc->xb.pdev->dev, "No serial number found.\n");
+		res = -EIO;
+		goto fail_exit;
+	}
+
+	/* Check for hardware echo cancel support */
+	if (!strncmp("TE133", wc->ddev->hardware_id+1, 5) ||
+	    !strncmp("TE134", wc->ddev->hardware_id+1, 5)) {
+		vpmcapable = true;
+	}
 
 	wc->wq = create_singlethread_workqueue(wc->name);
 	if (!wc->wq) {
@@ -2559,17 +2648,20 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	if (latency > 0 && latency < DRING_SIZE) {
 		wcxb_set_minlatency(&wc->xb, latency);
 		if (WCXB_DEFAULT_LATENCY != latency)
-			dev_info(&wc->dev->dev,
+			dev_info(&wc->xb.pdev->dev,
 				"latency manually overridden to %d\n",
 				latency);
 	} else {
-		dev_info(&wc->dev->dev,
+		dev_info(&wc->xb.pdev->dev,
 			"latency module parameter must be between 1 and %d\n",
 			DRING_SIZE);
 		res = -EPERM;
 		goto fail_exit;
 	}
 
+	wcxb_set_minlatency(&wc->xb, latency);
+	wcxb_set_maxlatency(&wc->xb, max_latency);
+
 	create_sysfs_files(wc);
 
 	res = t13x_hardware_post_init(wc, &type);
@@ -2582,10 +2674,15 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	if (res)
 		goto fail_exit;
 
-#ifdef VPM_SUPPORT
-	if (!wc->vpm)
+	if (!wc->vpm && vpmsupport && vpmcapable)
 		t13x_vpm_init(wc);
-#endif
+
+	if (wc->vpm)
+		wc->ddev->devicetype = kasprintf(GFP_KERNEL,
+				 "%s (VPMOCT032)", wc->devtype->name);
+	else
+		wc->ddev->devicetype = kasprintf(GFP_KERNEL, "%s",
+				wc->devtype->name);
 
 	wc->span.ops = &t13x_span_ops;
 	list_add_tail(&wc->span.device_node, &wc->ddev->spans);
@@ -2593,9 +2690,9 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	/* Span is in red alarm by default */
 	wc->span.alarms = DAHDI_ALARM_NONE;
 
-	res = dahdi_register_device(wc->ddev, &wc->dev->dev);
+	res = dahdi_register_device(wc->ddev, &wc->xb.pdev->dev);
 	if (res) {
-		dev_info(&wc->dev->dev, "Unable to register with DAHDI\n");
+		dev_info(&wc->xb.pdev->dev, "Unable to register with DAHDI\n");
 		goto fail_exit;
 	}
 
@@ -2603,10 +2700,10 @@ static int __devinit te13xp_init_one(struct pci_dev *pdev,
 	mod_timer(&wc->timer, jiffies + HZ/5);
 
 	if (wc->ddev->hardware_id) {
-		dev_info(&wc->dev->dev, "Found a %s (SN: %s)\n",
+		dev_info(&wc->xb.pdev->dev, "Found a %s (SN: %s)\n",
 				wc->devtype->name, wc->ddev->hardware_id);
 	} else {
-		dev_info(&wc->dev->dev, "Found a %s\n",
+		dev_info(&wc->xb.pdev->dev, "Found a %s\n",
 				wc->devtype->name);
 	}
 
@@ -2623,7 +2720,7 @@ fail_exit:
 static void __devexit te13xp_remove_one(struct pci_dev *pdev)
 {
 	struct t13x *wc = pci_get_drvdata(pdev);
-	dev_info(&wc->dev->dev, "Removing a Wildcard TE13xP.\n");
+	dev_info(&wc->xb.pdev->dev, "Removing a Wildcard TE13xP.\n");
 	if (!wc)
 		return;
 
@@ -2643,11 +2740,9 @@ static void __devexit te13xp_remove_one(struct pci_dev *pdev)
 	/* Turn off status LED */
 	t13x_setleds(wc, 0);
 
-#ifdef VPM_SUPPORT
 	if (wc->vpm)
 		release_vpm450m(wc->vpm);
 	wc->vpm = NULL;
-#endif
 
 	dahdi_unregister_device(wc->ddev);
 
@@ -2666,7 +2761,7 @@ static DEFINE_PCI_DEVICE_TABLE(te13xp_pci_tbl) = {
 static void te13xp_shutdown(struct pci_dev *pdev)
 {
 	struct t13x *wc = pci_get_drvdata(pdev);
-	dev_info(&wc->dev->dev, "Quiescing a Wildcard TE13xP.\n");
+	dev_info(&wc->xb.pdev->dev, "Quiescing a Wildcard TE13xP.\n");
 	if (!wc)
 		return;
 
@@ -2722,12 +2817,12 @@ module_param(alarmdebounce, int, S_IRUGO | S_IWUSR);
 module_param(losalarmdebounce, int, S_IRUGO | S_IWUSR);
 module_param(aisalarmdebounce, int, S_IRUGO | S_IWUSR);
 module_param(yelalarmdebounce, int, S_IRUGO | S_IWUSR);
-#ifdef VPM_SUPPORT
 module_param(vpmsupport, int, 0600);
-#endif
 module_param(force_firmware, int, S_IRUGO);
 module_param(latency, int, S_IRUGO);
 MODULE_PARM_DESC(latency, "How many milliseconds of audio to buffer between card and host (3ms default). This number will increase during runtime, dynamically, if dahdi detects that it is too small. This is commonly refered to as a \"latency bump\"");
+module_param(max_latency, int, 0600);
+MODULE_PARM_DESC(max_latency, "The maximum amount of latency that the driver will permit.");
 
 MODULE_DESCRIPTION("Wildcard Digital Card Driver");
 MODULE_AUTHOR("Digium Incorporated <support@digium.com>");
