@@ -29,6 +29,7 @@
 #include <linux/crc32.h>
 #include <linux/delay.h>
 #include <linux/version.h>
+#include <linux/slab.h>
 
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 26)
 #define HAVE_RATELIMIT
@@ -45,10 +46,13 @@
 
 /* The definition for Surprise Down was added in Linux 3.6 in (a0dee2e PCI: misc
  * pci_reg additions). It may be backported though so we won't check for the
- * version. */
+ * version. Same with PCI_EXP_SLTCTL_PDCE. */
 #ifndef PCI_ERR_UNC_SURPDN
 #define PCI_ERR_UNC_SURPDN 0x20
 #endif
+#ifndef PCI_EXP_SLTCTL_PDCE
+#define PCI_EXP_SLTCTL_PDCE 0x8
+#endif
 
 /* FPGA Status definitions */
 #define OCT_CPU_RESET		(1 << 0)
@@ -62,8 +66,9 @@
 #define DRING_SIZE_MASK		(DRING_SIZE-1)
 #define DESC_EOR		(1 << 0)
 #define DESC_INT		(1 << 1)
+#define DESC_IO_ERROR		(1 << 30)
 #define DESC_OWN		(1 << 31)
-#define DESC_DEFAULT_STATUS	0xdeadbeef
+#define DESC_DEFAULT_STATUS	0xdeadbe00
 #define DMA_CHAN_SIZE		128
 
 /* Echocan definitions */
@@ -351,16 +356,49 @@ static void _wcxb_reset_dring(struct wcxb *xb)
 	BUG_ON(!hdesc);
 	/* Set end of ring bit in last descriptor to force hw to loop around */
 	hdesc->control |= cpu_to_be32(DESC_EOR);
+#ifdef DEBUG
+	xb->last_retry_count = 0;
+#endif
 	iowrite32be(xb->hw_dring_phys, xb->membase + TDM_DRING_ADDR);
 }
 
 static void wcxb_handle_dma(struct wcxb *xb)
 {
+#ifdef DEBUG
+	bool did_retry_dma = false;
+	u8 retry;
+#endif
 	struct wcxb_meta_desc *mdesc;
+	struct wcxb_hw_desc *tail = &(xb->hw_dring[xb->dma_tail]);
 
-	while (!(xb->hw_dring[xb->dma_tail].control & cpu_to_be32(DESC_OWN))) {
+	while (!(tail->control & cpu_to_be32(DESC_OWN))) {
 		u_char *frame;
 
+		if (tail->control & cpu_to_be32(DESC_IO_ERROR)) {
+			u32 ier;
+			unsigned long flags;
+
+			/* The firmware detected an error condition on the bus.
+			 * Force an underrun by disabling the descriptor
+			 * complete interrupt. When the driver processes the
+			 * underrun it will reset the TDM engine. */
+			xb->flags.io_error = 1;
+
+			spin_lock_irqsave(&xb->lock, flags);
+			ier = ioread32be(xb->membase + IER);
+			iowrite32be(ier & ~DESC_COMPLETE, xb->membase + IER);
+			spin_unlock_irqrestore(&xb->lock, flags);
+			return;
+		}
+
+#ifdef DEBUG
+		retry = be32_to_cpu(tail->status) & 0xff;
+		if (xb->last_retry_count != retry) {
+			xb->last_retry_count = retry;
+			did_retry_dma = true;
+		}
+#endif
+
 		mdesc = &xb->meta_dring[xb->dma_tail];
 		frame = mdesc->rx_buf_virt;
 
@@ -368,6 +406,7 @@ static void wcxb_handle_dma(struct wcxb *xb)
 
 		xb->dma_tail =
 			(xb->dma_tail == xb->latency-1) ? 0 : xb->dma_tail + 1;
+		tail = &(xb->hw_dring[xb->dma_tail]);
 
 		mdesc = &xb->meta_dring[xb->dma_head];
 		frame = mdesc->tx_buf_virt;
@@ -379,6 +418,13 @@ static void wcxb_handle_dma(struct wcxb *xb)
 		xb->dma_head =
 			(xb->dma_head == xb->latency-1) ? 0 : xb->dma_head + 1;
 	}
+
+#ifdef DEBUG
+	if (did_retry_dma) {
+		dev_info(&xb->pdev->dev,
+			 "DMA retries detected: %d\n", xb->last_retry_count);
+	}
+#endif
 }
 
 static irqreturn_t _wcxb_isr(int irq, void *dev_id)
@@ -397,20 +443,40 @@ static irqreturn_t _wcxb_isr(int irq, void *dev_id)
 		if (pending & DESC_UNDERRUN) {
 			u32 reg;
 
-			/* bump latency */
+			/* Report the error in case drivers have any custom
+			 * methods for indicating potential data corruption. An
+			 * underrun means data loss in the TDM channel. */
+			if (xb->ops->handle_error)
+				xb->ops->handle_error(xb);
+
 			spin_lock(&xb->lock);
 
-			if (!xb->flags.latency_locked) {
-				xb->latency++;
+			if (xb->flags.io_error) {
+				/* Since an IO error is not necessarily because
+				 * the host could not keep up, we do not want to
+				 * bump the latency. */
+				xb->flags.io_error = 0;
+				dev_warn(&xb->pdev->dev,
+					 "IO error reported by firmware.\n");
+			} else if (!xb->flags.latency_locked) {
+				/* bump latency */
 
+				xb->latency = min(xb->latency + 1,
+						  xb->max_latency);
 #ifdef HAVE_RATELIMIT
 				if (__ratelimit(&_underrun_rl)) {
 #else
 				if (printk_ratelimit()) {
 #endif
-					dev_info(&xb->pdev->dev,
-						 "Underrun detected by hardware. Latency bumped to: %dms\n",
-						 xb->latency);
+					if (xb->latency != xb->max_latency) {
+						dev_info(&xb->pdev->dev,
+							 "Underrun detected by hardware. Latency bumped to: %dms\n",
+							 xb->latency);
+					} else {
+						dev_info(&xb->pdev->dev,
+							 "Underrun detected by hardware. Latency at max of %dms.\n",
+							 xb->latency);
+					}
 				}
 			}
 
@@ -540,27 +606,41 @@ static void _wcxb_pcie_hard_reset(struct wcxb *xb)
 {
 	struct pci_dev *const parent = xb->pdev->bus->self;
 	u32 aer_mask;
-	int pos;
+	u16 sltctl;
+	int pos_err;
+	int pos_exp;
 
 	if (!wcxb_is_pcie(xb))
 		return;
 
-	pos = pci_find_ext_capability(parent, PCI_EXT_CAP_ID_ERR);
-	if (pos) {
-		pci_read_config_dword(parent, pos + PCI_ERR_UNCOR_MASK,
+	pos_err = pci_find_ext_capability(parent, PCI_EXT_CAP_ID_ERR);
+	if (pos_err) {
+		pci_read_config_dword(parent, pos_err + PCI_ERR_UNCOR_MASK,
 				      &aer_mask);
-		pci_write_config_dword(parent, pos + PCI_ERR_UNCOR_MASK,
+		pci_write_config_dword(parent, pos_err + PCI_ERR_UNCOR_MASK,
 				       aer_mask | PCI_ERR_UNC_SURPDN);
 	}
 
+	/* Also disable any presence change reporting. */
+	pos_exp = pci_find_capability(parent, PCI_CAP_ID_EXP);
+	if (pos_exp) {
+		pci_read_config_word(parent, pos_exp + PCI_EXP_SLTCTL,
+				     &sltctl);
+		pci_write_config_word(parent, pos_exp + PCI_EXP_SLTCTL,
+				      sltctl & ~PCI_EXP_SLTCTL_PDCE);
+	}
+
 	_wcxb_hard_reset(xb);
 
-	if (pos) {
-		pci_write_config_dword(parent, pos + PCI_ERR_UNCOR_MASK,
+	if (pos_exp)
+		pci_write_config_word(parent, pos_exp + PCI_EXP_SLTCTL, sltctl);
+
+	if (pos_err) {
+		pci_write_config_dword(parent, pos_err + PCI_ERR_UNCOR_MASK,
 				       aer_mask);
 
 		/* Clear the error as well from the status register. */
-		pci_write_config_dword(parent, pos + PCI_ERR_UNCOR_STATUS,
+		pci_write_config_dword(parent, pos_err + PCI_ERR_UNCOR_STATUS,
 				       PCI_ERR_UNC_SURPDN);
 	}
 
@@ -602,6 +682,8 @@ int wcxb_init(struct wcxb *xb, const char *board_name, u32 int_mode)
 		return -EINVAL;
 
 	xb->latency = WCXB_DEFAULT_LATENCY;
+	xb->max_latency = WCXB_DEFAULT_MAXLATENCY;
+
 	spin_lock_init(&xb->lock);
 
 	xb->membase = pci_iomap(pdev, 0, 0);
@@ -778,7 +860,7 @@ struct wcxb_firm_header {
 	__le32	version;
 } __packed;
 
-static u32 wcxb_get_firmware_version(struct wcxb *xb)
+u32 wcxb_get_firmware_version(struct wcxb *xb)
 {
 	u32 version = 0;
 
@@ -793,7 +875,8 @@ static u32 wcxb_get_firmware_version(struct wcxb *xb)
 }
 
 static int wcxb_update_firmware(struct wcxb *xb, const struct firmware *fw,
-				const char *filename)
+				const char *filename,
+				enum wcxb_reset_option reset)
 {
 	u32 tdm_control;
 	static const int APPLICATION_ADDRESS = 0x200000;
@@ -845,14 +928,20 @@ static int wcxb_update_firmware(struct wcxb *xb, const struct firmware *fw,
 			 APPLICATION_ADDRESS + META_BLOCK_OFFSET,
 			 &meta, sizeof(meta));
 
-	/* Reset fpga after loading firmware */
-	dev_info(&xb->pdev->dev, "Firmware load complete. Reseting device.\n");
-	tdm_control = ioread32be(xb->membase + TDM_CONTROL);
+	if (WCXB_RESET_NOW == reset) {
+		/* Reset fpga after loading firmware */
+		dev_info(&xb->pdev->dev,
+				"Firmware load complete. Reseting device.\n");
+		tdm_control = ioread32be(xb->membase + TDM_CONTROL);
 
-	wcxb_hard_reset(xb);
+		wcxb_hard_reset(xb);
 
-	iowrite32be(0, xb->membase + 0x04);
-	iowrite32be(tdm_control, xb->membase + TDM_CONTROL);
+		iowrite32be(0, xb->membase + 0x04);
+		iowrite32be(tdm_control, xb->membase + TDM_CONTROL);
+	} else {
+		dev_info(&xb->pdev->dev,
+			"Delaying reset. Firmware load requires a power cycle\n");
+	}
 
 	wcxb_spi_device_destroy(flash_spi_device);
 	wcxb_spi_master_destroy(flash_spi_master);
@@ -860,10 +949,16 @@ static int wcxb_update_firmware(struct wcxb *xb, const struct firmware *fw,
 }
 
 int wcxb_check_firmware(struct wcxb *xb, const u32 expected_version,
-			const char *firmware_filename, bool force_firmware)
+			const char *firmware_filename, bool force_firmware,
+			enum wcxb_reset_option reset)
 {
 	const struct firmware *fw;
 	const struct wcxb_firm_header *header;
+	static const int APPLICATION_ADDRESS = 0x200000;
+	static const int META_BLOCK_OFFSET   = 0x170000;
+	struct wcxb_spi_master *flash_spi_master;
+	struct wcxb_spi_device *flash_spi_device;
+	struct wcxb_meta_block meta;
 	int res = 0;
 	u32 crc;
 	u32 version = 0;
@@ -882,6 +977,27 @@ int wcxb_check_firmware(struct wcxb *xb, const u32 expected_version,
 		return 0;
 	}
 
+	/* Check meta firmware version for a not-booted application image */
+	flash_spi_master = wcxb_spi_master_create(&xb->pdev->dev,
+						  xb->membase + FLASH_SPI_BASE,
+						  false);
+	flash_spi_device = wcxb_spi_device_create(flash_spi_master, 0);
+	res = wcxb_flash_read(flash_spi_device,
+			APPLICATION_ADDRESS + META_BLOCK_OFFSET,
+			&meta, sizeof(meta));
+	if (res) {
+		dev_info(&xb->pdev->dev, "Unable to read flash\n");
+		return -EIO;
+	}
+
+	if ((meta.version == cpu_to_le32(expected_version))
+			&& !force_firmware) {
+		dev_info(&xb->pdev->dev,
+			"Detected previous firmware updated to current version %x, but %x is currently running on card. You likely need to power cycle your system.\n",
+			expected_version, version);
+		return 0;
+	}
+
 	if (force_firmware) {
 		dev_info(&xb->pdev->dev,
 			"force_firmware module parameter is set. Forcing firmware load, regardless of version\n");
@@ -924,12 +1040,22 @@ int wcxb_check_firmware(struct wcxb *xb, const u32 expected_version,
 	dev_info(&xb->pdev->dev, "Found %s (version: %x) Preparing for flash\n",
 				firmware_filename, header->version);
 
-	res = wcxb_update_firmware(xb, fw, firmware_filename);
+	res = wcxb_update_firmware(xb, fw, firmware_filename, reset);
 
 	version = wcxb_get_firmware_version(xb);
-	dev_info(&xb->pdev->dev, "Reset into firmware version: %x\n", version);
+	if (WCXB_RESET_NOW == reset) {
+		dev_info(&xb->pdev->dev,
+			"Reset into firmware version: %x\n", version);
+	} else {
+		dev_info(&xb->pdev->dev,
+			"Running firmware version: %x\n", version);
+		dev_info(&xb->pdev->dev,
+			"Loaded firmware version: %x (Will load after next power cycle)\n",
+			header->version);
+	}
 
-	if ((expected_version != version) && !force_firmware) {
+	if ((WCXB_RESET_NOW == reset) && (expected_version != version)
+			&& !force_firmware) {
 		/* On the off chance that the interface is in a state where it
 		 * cannot boot into the updated firmware image, power cycling
 		 * the card can recover. A simple "reset" of the computer is not
