@@ -524,9 +524,12 @@ static void PRI_card_pcm_recompute(xpd_t *xpd, xpp_line_t pcm_mask)
 	//XPD_DBG(SIGNAL, xpd, "pcm_mask=0x%X\n", pcm_mask);
 	/* Add/remove all the trivial cases */
 	pcm_mask |= PHONEDEV(xpd).offhook_state;
+	if (priv->is_cas)
+		pcm_mask |= BITMASK(PHONEDEV(xpd).channels);
 	for_each_line(xpd, i)
 	    if (IS_SET(pcm_mask, i))
 		line_count++;
+	    else
 	if (priv->is_cas) {
 		if (priv->pri_protocol == PRI_PROTO_E1) {
 			/* CAS: Don't send PCM to D-Channel */
@@ -611,7 +614,7 @@ static int set_pri_proto(xpd_t *xpd, enum pri_protocol set_proto)
 	}
 	priv->pri_protocol = set_proto;
 	priv->is_cas = -1;
-	phonedev->channels = pri_num_channels(set_proto);
+	phonedev_alloc_channels(xpd, pri_num_channels(set_proto));
 	phonedev->offhook_state = BITMASK(phonedev->channels);
 	CALL_PHONE_METHOD(card_pcm_recompute, xpd, 0);
 	priv->deflaw = deflaw;
@@ -1082,6 +1085,7 @@ static int pri_set_spantype(struct dahdi_span *span, enum spantypes spantype)
 	struct phonedev *phonedev = container_of(span, struct phonedev, span);
 	xpd_t *xpd = container_of(phonedev, struct xpd, phonedev);
 	enum pri_protocol set_proto = PRI_PROTO_0;
+	int ret;
 
 	XPD_INFO(xpd, "%s: %s\n", __func__, dahdi_spantype2str(spantype));
 	switch (spantype) {
@@ -1099,7 +1103,13 @@ static int pri_set_spantype(struct dahdi_span *span, enum spantypes spantype)
 			__func__, dahdi_spantype2str(spantype));
 		return -EINVAL;
 	}
-	return set_pri_proto(xpd, set_proto);
+	ret = set_pri_proto(xpd, set_proto);
+	if (ret < 0) {
+		XPD_ERR(xpd, "%s: set_pri_proto failed\n", __func__);
+		return ret;
+	}
+	dahdi_init_span(span);
+	return 0;
 }
 
 static int PRI_card_open(xpd_t *xpd, lineno_t pos)
@@ -1224,7 +1234,7 @@ static int pri_chanconfig(struct file *file, struct dahdi_chan *chan,
 		if (priv->pri_protocol != PRI_PROTO_E1 && priv->is_cas != 0)
 			set_mode_cas(xpd, 0);
 	} else {
-		if (chan->channo == 1) {
+		if (chan->chanpos == 1) {
 			XPD_DBG(GENERAL, xpd,
 				"channel %d (%s) marked a not DChan\n",
 				chan->channo, chan->name);
@@ -1929,7 +1939,8 @@ static void PRI_card_pcm_tospan(xpd_t *xpd, xpacket_t *pack)
 				dchan_state(xpd, 0);
 		}
 		if (IS_SET(physical_mask, i)) {
-			r = XPD_CHAN(xpd, logical_chan)->readchunk;
+			struct dahdi_chan *chan = XPD_CHAN(xpd, logical_chan);
+			r = chan->readchunk;
 			// memset((u_char *)r, 0x5A, DAHDI_CHUNKSIZE);  // DEBUG
 			memcpy((u_char *)r, pcm, DAHDI_CHUNKSIZE);
 			pcm += DAHDI_CHUNKSIZE;
