@@ -425,6 +425,7 @@ struct dahdi_chan {
 	int lastnumbufs;
 #endif
 	spinlock_t lock;
+	struct mutex mutex;
 	char name[40];
 	/* Specified by DAHDI */
 	/*! \brief DAHDI channel number */
@@ -941,6 +942,7 @@ struct dahdi_device {
 	const char *devicetype;
 	struct device dev;
 	unsigned int irqmisses;
+	struct timespec registration_time;
 };
 
 struct dahdi_span {
@@ -1280,6 +1282,8 @@ static inline void dahdi_ec_span(struct dahdi_span *span)
 
 extern struct file_operations *dahdi_transcode_fops;
 
+int dahdi_get_auto_assign_spans(void);
+
 /* Don't use these directly -- they're not guaranteed to
    be there. */
 extern short __dahdi_mulaw[256];
@@ -1589,6 +1593,10 @@ struct mutex {
 #define	chan_err(chan, fmt, ...)	chan_printk(ERR, "", chan, fmt, \
 						## __VA_ARGS__)
 
+#ifndef pr_fmt
+#define pr_fmt(fmt)             KBUILD_MODNAME ": " fmt
+#endif
+
 #ifndef pr_err
 #define pr_err(fmt, ...) \
 	printk(KERN_ERR pr_fmt(fmt), ##__VA_ARGS__)
